// PiTrex I/O functions for initialising Raspberry Pi GPIO, and Reading/Writing from/to Vectrex
// Kevin Koster, 2019.
// V. 0.2.3

// PiTrex Discrete Read & Write Functions

//#include <bcm2835.h>
#include "bcm2835.h"
#include <unistd.h>

#include "pitrexio-gpio.h"
#include "via6522.h"

//Comment this out to poll RDY input instead of using Event Detection Register:
//Note: *_end functions may not work reliably without using the EDR.
//#define USE_EDR

//Set this to configure GPIO to use UART0 instead of UART1 on bits 14 and 15.
//This is the default config used by Linux when run on Raspberry Pis without WiFi support
//(but changing this is only required if the serial terminal is going to be used).
//#define UART0

/*
  Port Assignments:
  ^^^^^^^^^^^^^^^^^
Descrip.|Bit    |Pin    |
D0      |16     |36     |
D1      |17     |11     |
D2      |18     |12     |
D3      |19     |35     |
D4      |20     |38     |
D5      |21     |40     |
D6      |22     |15     |
D7      |23     |16     |

A0      |7      |26     |
A1      |8      |24     |
A2      |9      |21     |
A3      |10     |19     |
A12     |11     |23     |
A14     |12     |32     |
R/#W    |13     |33     |

PB6     |5      |29     |
IRQ_LATCH |6	|31		|

RDY(#OE)|24     |18     |
LATCH_EN|25     |22     |
IRQ     |26     |37     |
#HALT	|27	|13		| -Jumper to GND
#D/A    |N/A    |9      | -GND

3V3     |N/A    |1      |
5V      |N/A    |2      |
GND     |N/A    |6      |

Note: Pins counted from 1, bits counted from 0.
*/

//Begin read operation
//Returns 1 if address is valid, otherwise 0.
char vectrexread_begin (unsigned int address)
{
	unsigned char highaddressbits = 0;
	
	if ((address & 0x1000) == 0x1000)
	 highaddressbits += 0x10; //A12 set
	
	if ((address & 0xC000) == 0xC000)
	 highaddressbits += 0x20; //A15 and A14 set
	
	if ((highaddressbits == 0) || ((address & 0x2800) > 0) )
	 return 0; //6522 (VIA) not selected, or invalid address
	
	//Sets highest 2 bits per result of A15 AND A14, as well as A12 value. Result is ORed with lowest four bits of address.
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | highaddressbits) << 7;
	
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF);	//Set all outputs Low
	
	//Set Data bus to Input mode:
//For ALT0 (UART0) on 14-15, use 0x24249. For ALT5 (UART1), use 0x12249.
#ifdef UART0
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x24249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = INPUT
#else
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x12249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = INPUT
#endif
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8000);		// GPIO Pins 20-24 = INPUT | 25    = OUTPUT | 26-29 = INPUT
	
	bcm2835_gpio_set (13); // R/#W High
	// Set Address Bus
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), shortaddress); // Set Highs

#ifdef USE_EDR
	bcm2835_gpio_hen (24); // Detect when pin High
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop
	bcm2835_gpio_clr_hen (24); // Disable High detection
#else
	char status;
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status); // Loop until RDY goes High
#endif
	bcm2835_gpio_set (25); // LATCH EN High

#ifdef USE_EDR
	// Detect RDY rising edge with EDR:
	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
#endif
	return 1;
}

//Return read byte from previously started read operation.
//Returns after operation completed if it hasn't already by the time that it is called.
unsigned char vectrexread_end (void)
{
	unsigned char datain;
	char status;
#ifdef USE_EDR
	while (!bcm2835_gpio_eds (24)); //Loop
	bcm2835_gpio_clr_ren (24); // Disable rising edge detection
#else
	// Poll RDY for LOW to HIGH transistion:
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (status); // Loop until RDY goes Low
	
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status); // Loop until RDY goes High
#endif
	
	datain = ( bcm2835_peri_read (bcm2835_gpio + BCM2835_GPLEV0/4) >> 16) & 0xFF; // Read data byte
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFDFFF); //LATCH EN Low, IRQ LATCH Low, R/#W High
	return datain;
}

// Read a byte from the Vectrex at the specified address
unsigned char vectrexread (unsigned int address)
{
	if (vectrexread_begin (address))
	 return vectrexread_end ();
	else
	 return 0;
}


//Uses IRQ LATCH instead of LATCH EN output, causing the read to begin immediately after an IRQ signal is received from the 6522
//Begin read operation
//Returns 1 if address is valid, otherwise 0.
char vectrexread_begin_on_irq (unsigned int address)
{
	unsigned char highaddressbits = 0;
	
	if ((address & 0x1000) == 0x1000)
	 highaddressbits += 0x10; //A12 set
	
	if ((address & 0xC000) == 0xC000)
	 highaddressbits += 0x20; //A15 and A14 set
	
	if ((highaddressbits == 0) || ((address & 0x2800) > 0) )
	 return 0; //6522 (VIA) not selected, or invalid address
	
	//Sets highest 2 bits per result of A15 AND A14, as well as A12 value. Result is ORed with lowest four bits of address.
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | highaddressbits) << 7;
	
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF);	//Set all outputs Low
	
	//Set Data bus to Input mode:
//For ALT0 (UART0) on 14-15, use 0x24249. For ALT5 (UART1), use 0x12249.
#ifdef UART0
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x24249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = INPUT
#else
    bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x12249);       // GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = INPUT
#endif
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8000);		// GPIO Pins 20-24 = INPUT | 25    = OUTPUT | 26-29 = INPUT
	
	bcm2835_gpio_set (13); // R/#W High
	// Set Address Bus
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), shortaddress); // Set Highs

#ifdef USE_EDR
	bcm2835_gpio_hen (24); // Detect when pin High
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop
	bcm2835_gpio_clr_hen (24); // Disable High detection
#else
	char status;
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status); // Loop until RDY goes High
#endif
	bcm2835_gpio_set (6); // IRQ LATCH High

#ifdef USE_EDR
	// Detect RDY rising edge with EDR:
	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
#endif
	return 1;
}

// Read a byte from the Vectrex at the specified address after #IRQ asserted by the Vectrex
unsigned char vectrexread_on_irq (unsigned int address)
{
	if (vectrexread_begin_on_irq (address))
	 return vectrexread_end ();
	else
	 return 0;
}

// Begin writing a byte to the Vectrex
//Returns 1 if address is valid, otherwise 0.
char vectrexwrite_begin (unsigned int address, unsigned char data)
{
	unsigned char highaddressbits = 0;
	
	if ((address & 0x1000) == 0x1000)
	 highaddressbits += 0x10; //A12 set
	
	if ((address & 0xC000) == 0xC000)
	 highaddressbits += 0x20; //A15 and A14 set
	
	if ((highaddressbits == 0) || ((address & 0x2800) > 0) )
	 return 0; //6522 (VIA) not selected, or invalid address
	
	//Sets highest 2 bits per result of A15 AND A14, as well as A12 value. Result is ORed with lowest four bits of address.
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | highaddressbits) << 7;
	
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF); //LATCH EN Low  R/#W Low
	// Data Output Mode:
//For ALT0 (UART0) on 14-15, use 0x9264249. For ALT5 (UART1), use 0x9252249.
#ifdef UART0
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x9264249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = OUTPUT
#else
    bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x9252249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = OUTPUT
#endif
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8249);			// GPIO Pins 20-23 = OUTPUT | 24 = INPUT | 25    = OUTPUT | 26-29 = INPUT

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), (data << 16) | shortaddress ); // Data on Data Bus and Address on Address Bus

	bcm2835_gpio_set (25); //LATCH EN High
	
#ifdef USE_EDR
	// Detect RDY rising edge with EDR:
	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
#endif
	return 1;
}

// Detect whether write operation has completed, if so returns 1, else returns 0.
// If loop == 1, does not return until interrupt is detected.
char vectrexwrite_end (char loop)
{
	char status;
#ifdef USE_EDR
	do
	 {
	  status = bcm2835_gpio_eds (24);
	 }
	while (!status && loop); // Loop until RDY goes Low-to-High
	
	if (loop)
	 bcm2835_gpio_clr_ren (24); // Disable rising edge detection

#else
	// Poll RDY for LOW to HIGH transistion:
	if (loop)
	{
	 do
	 {
	  status = bcm2835_gpio_lev (24);
	 }
	 while (status); // Loop until RDY goes Low
	}
	
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status && loop); // Loop until RDY goes High
#endif
	
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF); //LATCH EN Low IRQ LATCH Low  R/#W Low
	bcm2835_gpio_set (13); // R/#W High
	return status;
}


// Write a byte to the Vectrex at the specified address.
// Returns after write is completed.
void vectrexwrite (unsigned int address, unsigned char data)
{
	if (vectrexwrite_begin (address, data))
	 vectrexwrite_end (1);
}

//Uses IRQ LATCH instead of LATCH EN output, causing the write to begin immediately after an IRQ signal is received from the 6522
// Begin writing a byte to the Vectrex
//Returns 1 if address is valid, otherwise 0.
char vectrexwrite_begin_on_irq (unsigned int address, unsigned char data)
{
	unsigned char highaddressbits = 0;
	
	if ((address & 0x1000) == 0x1000)
	 highaddressbits += 0x10; //A12 set
	
	if ((address & 0xC000) == 0xC000)
	 highaddressbits += 0x20; //A15 and A14 set
	
	if ((highaddressbits == 0) || ((address & 0x2800) > 0) )
	 return 0; //6522 (VIA) not selected, or invalid address
	
	//Sets highest 2 bits per result of A15 AND A14, as well as A12 value. Result is ORed with lowest four bits of address.
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | highaddressbits) << 7;
	
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF); //LATCH EN Low  R/#W Low
	// Data Output Mode:
//For ALT0 (UART0) on 14-15, use 0x9264249. For ALT5 (UART1), use 0x9252249.
#ifdef UART0
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x9264249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = OUTPUT
#else
    bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x9252249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = OUTPUT
#endif
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8249);			// GPIO Pins 20-23 = OUTPUT | 24 = INPUT | 25    = OUTPUT | 26-29 = INPUT

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), (data << 16) | shortaddress ); // Data on Data Bus and Address on Address Bus

	bcm2835_gpio_set (6); //IRQ LATCH High
	
#ifdef USE_EDR
	// Detect RDY rising edge with EDR:
	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
#endif
	return 1;
}

// Write a byte to the Vectrex at the specified address after #IRQ is asserted.
// Returns when written.
void vectrexwrite_on_irq (unsigned int address, unsigned char data)
{
	if (vectrexwrite_begin_on_irq (address, data))
	 vectrexwrite_end (1);
}

//Enable detection of Interrupt Request signal from the Vectrex
void vectrexinterrupt_begin (void)
{
#ifdef USE_EDR
	 bcm2835_gpio_hen (26); // Detect pin High
	 bcm2835_gpio_set_eds (26); //Clear Event Detection Register
#endif
}

/* Checks whether #IRQ signal was detected, if so returns 1, else returns 0.
*  If loop == 1, does not return until interrupt is detected.
*  Optionally reads Interrupt Flag Register, and only returns after specified interrupt is set.
*  If loop == 1, other interrupts than the one specified will be cleared automatically if they
*  are set to trigger #IRQ in the Interrupt Enable Register.
*
*  Values for "interrupt" are defined as VECTREX_VIA_IFR_* in pitrexio-gpio.h
*  If interrupt == 7 (VECTREX_VIA_IFR_IRQ_STATUS_FLAG), returns on any interrupt (doesn't check EFR).
*/
char vectrexinterrupt_end (char interrupt, char loop)
{
	char status = 0;
	unsigned char ier;

	if (interrupt != 7 && loop)
	 ier = vectrexread (VIA_int_enable);

	do
	{
	  if (interrupt != 7 && loop)
	   // Clear any other interrupts that are set in IER to trigger #IRQ (bits written "1" to IFR are cleared):
	   vectrexwrite (VIA_int_flags, ((0xFE << interrupt) | (0x7F >> (7 - interrupt))) & ier);
#ifdef USE_EDR
	 do
	  {
	   status = bcm2835_gpio_eds (26);
	   bcm2835_gpio_set_eds (26); //Clear Event Detection Register
	  }
	 while (!status && loop); // Loop until IRQ goes High
	 
	 if (loop)
	  bcm2835_gpio_clr_ren (24); // Disable rising edge detection
#else
	 do
	 {
	  status = bcm2835_gpio_lev (26);
	 }
	 while (!status && loop); // Loop until IRQ goes High
#endif
	}
	while (interrupt != 7 && !(vectrexread(VIA_int_flags) & (1 << interrupt)) && loop );
#ifdef USE_EDR
	bcm2835_gpio_clr_hen (26); // Disable High detection
#endif
	return status; // status == 1 if interrupt request signal received, else 0.
}
		

// Wait on interrupt from Vectrex
// Returns when Interrupt Request signal (#IRQ) detected as asserted.
void vectrexinterrupt (char interrupt)
{
	vectrexinterrupt_begin ();
	vectrexinterrupt_end (interrupt, 1);
}

// Initialise GPIO for communication with the Vectrex
// Vectrex can be initialised after the GPIO if viaconfig == 1,
// which sets it to the same initialisation state as the Vectrex BIOS would.
// With viaconfig == 0, the VIA and PSG registers are unchanged.
int vectrexinit (char viaconfig)
{
	unsigned char psgaddress;
	
//	bcm2835_debug(1); //Set debugging mode (GPIO not set, but actions reported)
	int result = bcm2835_init();
	if (result)
	{
	 //Set Initial State (ideally set to configured defaults at power-on):
	 //NOTE: This will disable other functions of these GPIO pins (except for the serial terminal on the UART)
	  //Input/Output Modes:
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL0/4), 0x9240000);	// GPIO Pins 0-5   = INPUT | 6-9   = OUTPUT
//For ALT0 (UART0) on 14-15, use 0x24249. For ALT5 (UART1), use 0x12249.
#ifdef UART0
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x12249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = INPUT
#else
     bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x24249);
#endif
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8000);		// GPIO Pins 20-24 = INPUT | 25    = OUTPUT | 26-29 = INPUT
	  //Output Levels:
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFDFFF);	//Set all outputs Low except R/#W
	 
	 bcm2835_gpio_set_pud (27, BCM2835_GPIO_PUD_UP);  // Enable Pull-Up on #HALT
	 
	 bcm2835_gpio_set_pud (24, BCM2835_GPIO_PUD_OFF); // Disable PU/D on RDY
//	 bcm2835_gpio_set_pud (25, BCM2835_GPIO_PUD_OFF); // and LATCH EN
//	 bcm2835_gpio_set_pud (6, BCM2835_GPIO_PUD_OFF);  // and IRQ LATCH

	 bcm2835_gpio_set (13); // R/#W High
	 
	 //Check that Vectrex CPU is halted.
	 if (bcm2835_gpio_lev (27))
	  result = VECTREX_HALT_NOT_ASSERTED;
	 else
	 {
	  //Initialise 6522 (VIA) and AY3-8912 (PSG) in standard Vectrex start-up state (with reference to Vectrex BIOS "INITALL" routine):
	  if (viaconfig == 1)
	  {
	   vectrexwrite (VIA_DDR_b, 0x9F); // All Outputs except COMPARE input and PB6 at Cartridge Port
	   vectrexwrite (VIA_DDR_a, 0xFF); // All Outputs
	   vectrexwrite (VIA_aux_cntl, 0x98); //Shift Reg. Enabled, T1 PB7 Enabled
	   vectrexwrite (VIA_t1_cnt_lo, 0x7F); // default scale = 0x7f
	   vectrexwrite (VIA_t2_lo, 0x30);
	   vectrexwrite (VIA_t2_hi, 0x75); // T2 set to Refresh = MSEC20. Reset T2 Interrupt Flag and start Frame Timer.
 
	   // Clear Programmable Sound Generator (PSG) registers
	   //  Should be done by the hardware reset signal anyway. But monkey see, monkey do.
	   for (psgaddress = 0x0E; psgaddress > 0; psgaddress--)
	   {
	    vectrexwrite (VIA_port_a, psgaddress); // Address
	    vectrexwrite (VIA_port_b, 0x19);
	    vectrexwrite (VIA_port_b, 1); // PSG Address Latch
	    vectrexwrite (VIA_port_a, 0); // Data
	    vectrexwrite (VIA_port_b, 0x11);
	    vectrexwrite (VIA_port_b, 1); // PSG Data Latch
	   }
	  }
	  //Add alternative initialisation states here?
	 }
	}

	return result;
}

int vectrexclose (void)
{
	int result = bcm2835_close();
	return result;
}
