// PiTrex I/O functions for initialising Raspberry Pi GPIO, and Reading/Writing from/to Vectrex
// Kevin Koster, 2019.
// V. 0.1

// PiTrex Discrete Read & Write Functions

#include <stddef.h> // these are fresstanding includes!
#include <stdint.h> // also "available":  <float.h>, <iso646.h>, <limits.h>, <stdarg.h>
#include <stdio.h>

#include "bcm2835.h"
#include "pitrexio-gpio.h"

//Comment this out to poll RDY input instead of using Event Detection Register:
//#define USE_EDR
/*UART1 USE */
#define GPIO1_PERI_INPUT 0x12249
#define GPIO1_PERI_OUTPUT 0x9252249


/* UART1 NOT USE 
#define GPIO1_PERI_INPUT 0x24249
#define GPIO1_PERI_OUTPUT 0x9264249
*/

/*
  Port Assignments:
  ^^^^^^^^^^^^^^^^^
Descrip.|Bit    |Pin    |
D0      |16     |36     |
D1      |17     |11     |
D2      |18     |12     |
D3      |19     |35     |
D4      |20     |38     |
D5      |21     |40     |
D6      |22     |15     |
D7      |23     |16     |

A0      |7      |26     |
A1      |8      |24     |
A2      |9      |21     |
A3      |10     |19     |
A12     |11     |23     |
A14     |12     |32     |
R/#W    |13     |33     |

PB6     |5      |29     |
IRQ_LATCH |6	|31		|

RDY(#OE)|24     |18     |
LATCH_EN|25     |22     |
IRQ     |26     |37     |
#HALT	|27	|13		| -Jumper to GND
#D/A    |N/A    |9      | -GND

3V3     |N/A    |1      |
5V      |N/A    |2      |
GND     |N/A    |6      |

Note: Pins counted from 1, bits counted from 0.
*/


// Read a byte from the Vectrex at the specified address
unsigned char vectrexread (unsigned int address)
{
	unsigned char datain;
	unsigned char highaddressbits = 0;

	if ((address & 0x1000) == 0x1000)
	 highaddressbits += 0x10; //A12 set

	if ((address & 0xC000) == 0xC000)
	 highaddressbits += 0x20; //A15 and A14 set

	if ((highaddressbits == 0) || ((address & 0x2800) > 0) )
	 return 0; //6522 (VIA) not selected, or invalid address

	//Sets highest 2 bits per result of A15 AND A14, as well as A12 value. Result is ORed with lowest four bits of address.
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | highaddressbits) << 7;

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF);	//Set all outputs Low

	//Set Data bus to Input mode:
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), GPIO1_PERI_INPUT);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = INPUT
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8000);		// GPIO Pins 20-24 = INPUT | 25    = OUTPUT | 26-29 = INPUT

	bcm2835_gpio_set (13); // R/#W High
	// Set Address Bus
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), shortaddress); // Set Highs

#ifdef USE_EDR
	bcm2835_gpio_hen (24); // Detect when pin High
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop - optimise? - Should use interrupts.
	bcm2835_gpio_clr_hen (24); // Disable High detection
#else
	char status;
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status); // Loop until RDY goes High
#endif
	bcm2835_gpio_set (25); // LATCH EN High

#ifdef USE_EDR
	// Detect RDY rising edge with EDR:
	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop - optimise? - Should use interrupts.
	bcm2835_gpio_clr_ren (24); // Disable rising edge detection
#else
	// Poll RDY for LOW to HIGH transistion:
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (status); // Loop until RDY goes Low

	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status); // Loop until RDY goes High
#endif

	datain = ( bcm2835_peri_read (bcm2835_gpio + BCM2835_GPLEV0/4) >> 16) & 0xFF; // Read data byte
	bcm2835_gpio_clr (25); // LATCH EN Low
	return datain;
}


//Uses IRQ LATCH instead of LATCH EN output, causing the read to begin immediately after an IRQ signal is received from the 6522
unsigned char vectrexread_on_irq (unsigned int address)
{
	unsigned char datain;
	unsigned char highaddressbits = 0;

	if ((address & 0x1000) == 0x1000)
	 highaddressbits += 0x10; //A12 set

	if ((address & 0xC000) == 0xC000)
	 highaddressbits += 0x20; //A15 and A14 set

	if ((highaddressbits == 0) || ((address & 0x2800) > 0) )
	 return 0; //6522 (VIA) not selected, or invalid address

	//Sets highest 2 bits per result of A15 AND A14, as well as A12 value. Result is ORed with lowest four bits of address.
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | highaddressbits) << 7;

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF);	//Set all outputs Low

	//Set Data bus to Input mode:
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), GPIO1_PERI_INPUT);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = INPUT
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8000);		// GPIO Pins 20-24 = INPUT | 25    = OUTPUT | 26-29 = INPUT

	bcm2835_gpio_set (13); // R/#W High
	// Set Address Bus
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), shortaddress); // Set Highs

#ifdef USE_EDR
	bcm2835_gpio_hen (24); // Detect when pin High
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop - optimise? - Should use interrupts.
	bcm2835_gpio_clr_hen (24); // Disable High detection
#else
	char status;
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status); // Loop until RDY goes High
#endif
	bcm2835_gpio_set (6); // IRQ LATCH High

#ifdef USE_EDR
	// Detect RDY rising edge with EDR:
	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop - optimise? - Should use interrupts.
	bcm2835_gpio_clr_ren (24); // Disable rising edge detection
#else
	// Poll RDY for LOW to HIGH transistion:
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (status); // Loop until RDY goes Low

	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status); // Loop until RDY goes High
#endif

	datain = ( bcm2835_peri_read (bcm2835_gpio + BCM2835_GPLEV0/4) >> 16) & 0xFF; // Read data byte
	bcm2835_gpio_clr (6); // IRQ LATCH Low
	return datain;
}

// Write a byte to the Vectrex at the specified address
void vectrexwrite (unsigned int address, unsigned char data)
{
	unsigned char highaddressbits = 0;

	if ((address & 0x1000) == 0x1000)
	 highaddressbits += 0x10; //A12 set

	if ((address & 0xC000) == 0xC000)
	 highaddressbits += 0x20; //A15 and A14 set

	if ((highaddressbits == 0) || ((address & 0x2800) > 0) )
	 return; //6522 (VIA) not selected, or invalid address

	//Sets highest 2 bits per result of A15 AND A14, as well as A12 value. Result is ORed with lowest four bits of address.
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | highaddressbits) << 7;

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF); //LATCH EN Low  R/#W Low
	// Data Output Mode:
	
//xx00 1001 0010 0101 0010 0010 0100 1001
//09252249
//000 = input
//001 = output
//010 = alt 5
	
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), GPIO1_PERI_OUTPUT);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = OUTPUT
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8249);			// GPIO Pins 20-23 = OUTPUT | 24 = INPUT | 25    = OUTPUT | 26-29 = INPUT

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), (data << 16) | shortaddress ); // Data on Data Bus and Address on Address Bus

	bcm2835_gpio_set (25); //LATCH EN High

#ifdef USE_EDR

	// Detect RDY rising edge with EDR:
	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop - optimise? - Should use interrupts.
	bcm2835_gpio_clr_ren (24); // Disable rising edge detection

#else
	// Poll RDY for LOW to HIGH transistion:
	char status;
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (status); // Loop until RDY goes Low

	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status); // Loop until RDY goes High
#endif

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF); //LATCH EN Low  R/#W Low
	bcm2835_gpio_set (13); // R/#W High
}

//Uses IRQ LATCH instead of LATCH EN output, causing the write to begin immediately after an IRQ signal is received from the 6522
void vectrexwrite_on_irq (unsigned int address, unsigned char data)
{
	unsigned char highaddressbits = 0;

	if ((address & 0x1000) == 0x1000)
	 highaddressbits += 0x10; //A12 set

	if ((address & 0xC000) == 0xC000)
	 highaddressbits += 0x20; //A15 and A14 set

	if ((highaddressbits == 0) || ((address & 0x2800) > 0) )
	 return; //6522 (VIA) not selected, or invalid address

	//Sets highest 2 bits per result of A15 AND A14, as well as A12 value. Result is ORed with lowest four bits of address.
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | highaddressbits) << 7;

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF); //IRQ LATCH Low, LATCH EN Low,  R/#W Low
	// Data Output Mode:
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), GPIO1_PERI_OUTPUT);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = OUTPUT
//	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x9264249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = OUTPUT
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8249);			// GPIO Pins 20-23 = OUTPUT | 24 = INPUT | 25    = OUTPUT | 26-29 = INPUT

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), (data << 16) | shortaddress ); // Data on Data Bus and Address on Address Bus

	bcm2835_gpio_set (6); //IRQ LATCH High

#ifdef USE_EDR

	// Detect RDY rising edge with EDR:
	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop - optimise? - Should use interrupts.
	bcm2835_gpio_clr_ren (24); // Disable rising edge detection

#else
	// Poll RDY for LOW to HIGH transistion:
	char status;
	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (status); // Loop until RDY goes Low

	do
	{
	 status = bcm2835_gpio_lev (24);
	}
	while (!status); // Loop until RDY goes High
#endif

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF); //IRQ LATCH Low  R/#W Low
	bcm2835_gpio_set (13); // R/#W High
}

// Wait until interrupt from Vectrex
void vectrexinterrupt (char interrupt)
{
	if (interrupt)
	{
#ifdef USE_EDR

	 bcm2835_gpio_hen (26); // Detect pin High
	 bcm2835_gpio_set_eds (26); //Clear Event Detection Register
	 while (!bcm2835_gpio_eds (26)); //Loop - optimise? - Should use interrupts.
	 bcm2835_gpio_clr_hen (26); // Disable High detection
#else
	char status;
	do
	{
	 status = bcm2835_gpio_lev (26);
	}
	while (!status); // Loop until RDY goes High
#endif
	}
}

int vectrexclose ()
{
	int result = bcm2835_close();
	return result;
}

// Initialise GPIO for communication with the Vectrex
int vectrexinit (char viaconfig)
{
    printf("vectrexinit()\r\n");

  
	//bcm2835_debug(1); //Set debugging mode (GPIO not set, but actions reported)
	int result = bcm2835_init();
//	int i;
	if (result)
	{
	 //Set Initial State (ideally set to configured defaults at power-on):
	 //NOTE: This will disable other functions of these GPIO pins
	  //Input/Output Modes:
//	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL0/4), 0x9240000);	// GPIO Pins 0-5   = INPUT | 6-9   = OUTPUT
	 
	 bcm2835_gpio_fsel(0, BCM2835_GPIO_FSEL_INPT);
	 bcm2835_gpio_fsel(1, BCM2835_GPIO_FSEL_INPT);
	 bcm2835_gpio_fsel(2, BCM2835_GPIO_FSEL_INPT);
	 bcm2835_gpio_fsel(3, BCM2835_GPIO_FSEL_INPT);
	 bcm2835_gpio_fsel(4, BCM2835_GPIO_FSEL_INPT);
	 bcm2835_gpio_fsel(5, BCM2835_GPIO_FSEL_INPT);
	 bcm2835_gpio_fsel(6, BCM2835_GPIO_FSEL_OUTP);
	 bcm2835_gpio_fsel(7, BCM2835_GPIO_FSEL_OUTP);
	 bcm2835_gpio_fsel(8, BCM2835_GPIO_FSEL_OUTP);
	 bcm2835_gpio_fsel(9, BCM2835_GPIO_FSEL_OUTP);
	 
//xx00 0000 0000 0001 0010 0010 0100 1001
//   12249  

	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), GPIO1_PERI_INPUT);//0x24249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = TTY | 16-19 = INPUT
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8000);		// GPIO Pins 20-24 = INPUT | 25    = OUTPUT | 26-29 = INPUT

	 
	 //Output Levels:
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF);	//Set all outputs Low
	 bcm2835_gpio_set_pud (27, BCM2835_GPIO_PUD_UP);  // Enable Pull-Up on #HALT
	 bcm2835_gpio_set_pud (24, BCM2835_GPIO_PUD_OFF); // Disable PU/D on RDY
	 bcm2835_gpio_set_pud (25, BCM2835_GPIO_PUD_OFF); // and LATCH EN
	 bcm2835_gpio_set_pud (6, BCM2835_GPIO_PUD_OFF);  // and IRQ LATCH
/*
	 for (i=7; i < 14 ; i++)
	  bcm2835_gpio_set_pud (i, BCM2835_GPIO_PUD_OFF); //Disable Pull-up/down on GPIO bits 7-13
	 for (i=16 ; i < 27; i++)
	  bcm2835_gpio_set_pud (i, BCM2835_GPIO_PUD_OFF); //Disable Pull-up/down on GPIO bits 16-26
*/
	 bcm2835_gpio_set (13); // R/#W High

	 //Check that Vectrex CPU is halted.
	 if (bcm2835_gpio_lev (27))
	  result = VECTREX_HALT_NOT_ASSERTED;

	 //Initialise 6522 port configuration in standard Vectrex start-up state:
	 if (viaconfig)
	 {
#define VIA_DDR_b     ((unsigned short) 0xD002) // VIA port B data direction register (0=input 1=output)
#define VIA_DDR_a     ((unsigned short) 0xD003) // VIA port A data direction register (0=input 1=output)
#define VIA_aux_cntl  ((unsigned short) 0xD00B) // VIA auxiliary control
	  vectrexwrite (VIA_DDR_b, 0x9F); // All Outputs except COMPARE input and PB6 at Cartridge Port
	  vectrexwrite (VIA_DDR_a, 0xFF); // All Outputs
	  vectrexwrite (VIA_aux_cntl, 0x98); //Shift Reg. Enabled, T1 PB7 Enabled
	 }

	}
    printf("vectrexinit() - done\r\n");

	return result;
}
